
/* UnQuill: Disassemble games written with the "Quill" adventure game system
    Copyright (C) 1996-2023  John Elliott

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include "unquill.h"

#ifndef TINY

extern const char *separator;

uchr qnextbyte_uc(ushrt *n)
{
	return (255-zmem((*n)++));
}

uchr qnextbyte_dict(ushrt *n)
{
	return zmem( (*n)++) & 0x7F;
}

uchr (*qnextbyte)(ushrt *) = qnextbyte_uc;

static char qwordch(char c)
{
	if (arch != ARCH_C64) return(0xFF - c);
	else                  return((0xFF - c)  & 0x7F);
}

void qlistconn(ushrt base, uchr num)
{
	ushrt n, addr;

	fprintf(outfile, "/CON\t\t;Connections\n");
	for (n = 0; n < num; n++)
	{
		fprintf(outfile, "/%d\n", n);

		addr = zword(base + 2 * n);
		while (zmem(addr) != 0xFF)
		{
			fprintf(outfile, "\t%s\t%d\n",
				qfindword(zmem(addr)),
				zmem(addr+1));
			addr += 2;
		}

	}	
}

void qlistcond(int type, ushrt first)
{
	uchr noun, verb;
	ushrt addr=first, subsaddr;

	fprintf(outfile, "/PRO\t%d\n", type);
	while (zmem(addr))
	{
		verb = zmem(addr++);
		noun = zmem(addr++);
		fprintf(outfile, "%s\t", (verb == 0xFF) ? "_" : qfindword(verb));
		fprintf(outfile, "%s\t", (noun == 0xFF) ? "_" : qfindword(noun));
		subsaddr = zword(addr++);
		while (zmem(subsaddr) != 0xFF)
		{
			qopcond(&subsaddr);
			fprintf(outfile, "\n\t\t");
		}
		subsaddr++;
		while (zmem(subsaddr) != 0xFF)
		{
			qopact(&subsaddr);
			if (zmem(subsaddr) != 0xFF) fprintf(outfile, "\n\t\t");
		}
		fprintf(outfile, "\n\n");
		addr++;
	}
	fprintf(outfile, "%s\n", separator);
}

void qopcond(ushrt *condad)
{
	uchr id=zmem((*condad)++);
	uchr param1 = zmem((*condad)++);

	opcact(id, "AT     NOTAT  ATGT   ATLT   PRESENTABSENT WORN   NOTWORN"
             "CARRIEDNOTCARRCHANCE ZERO   NOTZEROEQ     GT     LT     ");

	fprintf(outfile,"%4d ",param1);
	if (id > 12) 
	{
		fprintf(outfile,"%4d ",zmem((*condad)++));
	}
}


void qopact(ushrt *actad)
{
	static char inited=0;
	static char params[40];
	static char ptype[40];
	int n;
	uchr id=zmem((*actad)++);
	uchr actpar[2];

	if (!inited)
	{
		if (dbver > 0 && dbver < 10)
		{
			for (n = 0; n < 39; n++) 
				{
				if (n < 13)      params[n] = 0;
				else if (n < 29) params[n] = 1;
				else             params[n] = 2;

				if (n < 19)      ptype[n] = NIL;
				else if (n < 27) ptype[n] = OBJ;
				else             ptype[n] = NIL;
			}
			params[25] = params[26] = 2;
			ptype[17] = LOC;
			ptype[18] = MSG;
			ptype[25] = SWAP;
			ptype[26] = PLC;
		}
	else if (dbver >= 10)
  	{
		 for(n=0;n<39;n++)
   		 {
			if      (n < 17) params[n] = 0;
			else if (n < 33) params[n] = 1;
			else             params[n] = 2;

			if      (n < 21) ptype[n] = NIL;
			else if (n < 29) ptype[n] = OBJ;
			else             ptype[n] = NIL;
		}	
		params[29] = params[30] = 2;
		ptype[21]=LOC;
		ptype[22]=MSG;
		ptype[29]=SWAP;
		ptype[30]=PLC;

		if (arch == ARCH_CPC) params[19] = 2;	/* On CPC, INK takes 2 parameters */
	}
	else 
	{
		for(n = 0; n < 39; n++)
    		{
			if      (n < 11) params[n] = 0;
			else if (n < 23) params[n] = 1;
			else             params[n] = 2;
			if 	(n < 12) ptype[n]=NIL;
			else if (n < 21) ptype[n]=OBJ;
			else             ptype[n]=NIL;
		}
		params[20] = 2;
		ptype[12] = LOC;
		ptype[13] = MSG;
		ptype[20] = SWAP;
	}
	inited = 1;
  }

if (dbver > 0 && dbver < 10)
	   opcact(id,"INVEN  DESC   QUIT   END    DONE   OK     ANYKEY "
                     "SAVE   LOAD   TURNS  SCORE  CLS    DROPALLPAUSE  "
                     "PAPER  INK    BORDER GOTO   MESSAGEREMOVE GET    "
                     "DROP   WEAR   DESTROYCREATE SWAP   PLACE  SET    "
                     "CLEAR  PLUS   MINUS  LET    BEEP   ");


else if (dbver >= 10)
	   opcact(id,"INVEN  DESC   QUIT   END    DONE   OK     ANYKEY "
                     "SAVE   LOAD   TURNS  SCORE  CLS    DROPALLAUTOG  "
                     "AUTOD  AUTOW  AUTOR  PAUSE  PAPER  INK    BORDER "
                     "GOTO   MESSAGEREMOVE GET    DROP   WEAR   DESTROY"
                     "CREATE SWAP   PLACE  SET    CLEAR  PLUS   MINUS  "
                     "LET    BEEP   ");

else       opcact(id,"INVEN  DESC   QUIT   END    DONE   OK     ANYKEY "
                     "SAVE   LOAD   TURNS  SCORE  PAUSE  GOTO   MESSAGE"
                     "REMOVE GET    DROP   WEAR   DESTROYCREATE SWAP   "
                     "SET    CLEAR  PLUS   MINUS  LET    BEEP   ");

	if (id < 40) for (n=params[id]; n>0;)
	{
		actpar[n-1]=(zmem((*actad)++));
		fprintf(outfile,"%4d ",actpar[--n]);
	}


}





char *qword(ushrt addr)
{
	static char buf[50];
	short n;

	buf[0] = 0;
	for (n = 0; n < 4; n++)
	{
		buf[n] = qwordch(zmem(addr++)); 
	}
	buf[n] = 0;
	return buf;
}


char *qfindword(uchr w)
{
	int n;
	static char buf[10];

	for (n=vocab; zmem(n); n=n+5)
	{
		if (zmem(n+4) == w) return qword(n);
	}
	sprintf(buf, "(%d)", w);
	return buf;	
}


void qlistwords(ushrt first)
{
	ushrt n, m;

	fprintf(outfile, "/VOC\t\t;Vocabulary\n");
	fprintf(outfile, ";\t\tMovements are numbered < 13\n");
  
	for (n=first; zmem(n); n=n+5)
	{
		for (m = 0; m < 4; m++) 
		{
			fputc(qwordch(zmem(n + m)), outfile);
		}
		fprintf(outfile, "\t%d\n", zmem(n + 4));
	}
	fprintf(outfile, "%s\n", separator);
}

void qlistpos(ushrt postab, ushrt objmap, uchr num)
{
	ushrt n;
	uchr v;

	fprintf(outfile, "/OBJ\t\t;Object definitions\n");
	fprintf(outfile, ";obj\tstarts\tword\n");
	fprintf(outfile, ";num\t  at  \t\n");

	for(n = 0; n < num; n++)
	{
		fprintf(outfile, "/%d\t", n);
		v = zmem(postab+n);
		switch (v)
		{
			case 252: fprintf(outfile, "_"); break;	/* Does not exist */
			case 253: fprintf(outfile, "WORN"); break;
			case 254: fprintf(outfile, "CARRIED"); break;
			default: fprintf(outfile, "%d", v);
		}
		fprintf(outfile, "\t");
		if (dbver >= 10)
		{
			v = zmem(objmap + n);
			if (v != 0xFF)
				fprintf(outfile, "%s", qfindword(v));
			else	fputc('_', outfile);
		}
		else fputc('_', outfile);
		fputc('\n', outfile);
	}
	fprintf(outfile, "%s\n", separator);
}


void qlistmsgs(char *section, ushrt mtab, uchr num)
{
	ushrt n;
	uchr cch, term;
	ushrt addr;

	if (arch == ARCH_SPECTRUM) term = 0x1F; else term = 0;
	cch = ~term;

	fprintf(outfile, "%s\n", section);
	for(n = 0; n < num; n++)
	{
		fprintf(outfile, "/%d\n", n);
		addr = zword(mtab + 2*n);
		xpos=0;

		while (1)
		{
		    cch=(0xFF-(zmem(addr++)));
			if (cch == term) break;
		    qexpch(cch,&addr);
		}
		fprintf(outfile, "\n");
	}
	fprintf(outfile, "%s\n", separator);
}

void qlistsys(char *section, ushrt addr, uchr num)
{
	ushrt n;
	uchr cch, term;

	fprintf(outfile, "%s\n", section);
	if (arch == ARCH_SPECTRUM) term = 0x1F; else term = 0;
	cch = ~term;

	for(n = 0; n < num;)
	{
		fprintf(outfile, "/%d\n", n);
		xpos=0;

		while (1)
		{
		    cch=(0xFF-(zmem(addr++)));
			if (cch == term) break;
		    qexpch(cch,&addr);
		}
		fprintf(outfile, "\n");
		++n;
	}
	fprintf(outfile, "%s\n", separator);
}

#if 0


void qoneitem(ushrt table, uchr item)
{
	ushrt n;
	uchr  cch;
	uchr  term;

	if (arch == ARCH_SPECTRUM) term = 0x1F; else term = 0;
	cch = ~term;

	fprintf(outfile, "      <text>");	
	n=zword(table+2*item);
	xpos=0;


	while (1)
	{
	    cch=(0xFF-(zmem(n++)));
		if (cch == term) break;
	    qexpch(cch,&n);
	}
	fprintf(outfile, "</text>\n");	
}

#endif

void qexpdict(uchr cch)
{
	ushrt d=dict;
	qnextbyte = qnextbyte_dict;

 	if (dbver > 0) /* Early games aren't compressed & have no dictionary */
	{
		cch -= 164;
		while (cch)
		{
       			if (zmem(d++) & 0x80) cch--;
		}

      /* d=address of expansion text */

		while (1)
		{
			qexpch (zmem(d++) & 0x7F, &d);
			--d;
			if (zmem(d) & 0x80) break;
			++d;
		}
	}
	qnextbyte = qnextbyte_uc;
}



void qexpch_c64(uchr cch, ushrt *n)
{
	if (cch >= 'A' && cch <= 'Z') cch = tolower(cch);
	cch &= 0x7F;

	switch(cch)
	{	
		case 0x0D:
		fprintf (outfile,"\n");
		xpos = 0;
		break;

		case '\\':fprintf (outfile,"\\\\");    ++xpos; break;
		case ';':
		case '/': if (xpos == 0) fputc('\\', outfile);
			  fputc(cch, outfile);
			  ++xpos;
		 	  break;
		default:
		if (cch >= ' ' && cch < 127)
			fputc(cch, outfile);
		else	fprintf(outfile, "\\0x%02x", cch);
		++xpos;
		if (xpos >= 40)
		{
			fprintf(outfile,"\n");
			xpos = 0;
		}
		break;
	}
}



void qexpch_cpc(uchr cch, ushrt *n)
{
	switch(cch)
	{
		case 0x0d:
		case 0x14: 
		          xpos = 0;
			  fprintf (outfile,"\n"); break;	
		case '\\':fprintf (outfile,"\\\\");    ++xpos; break;
		case ';':
		case '/': if (xpos == 0) fputc('\\', outfile);
			  fputc(cch, outfile);
			  ++xpos;
		 	  break;

		default:
			if (cch >= ' ' && cch < 127)
				fputc(cch, outfile);
			else	fprintf(outfile, "\\0x%02x", cch);
		++xpos;
		if (xpos >= 40)
		{
			fprintf(outfile,"\n");
			xpos = 0;
		}
		break;
	}
}

#if 0
static const ushrt zx_entity[] = 
{
	0x00A0, /*	NO-BREAK SPACE */
	0x259D, /*	QUADRANT UPPER RIGHT */
	0x2598, /*	QUADRANT UPPER LEFT */
	0x2580, /*	UPPER HALF BLOCK */
	0x2597, /*	QUADRANT LOWER RIGHT */
	0x2590, /*	RIGHT HALF BLOCK */
	0x259A, /*	QUADRANT UPPER LEFT AND LOWER RIGHT */
	0x259C, /*	QUADRANT UPPER LEFT AND UPPER RIGHT AND LOWER RIGHT */
	0x2596, /*	QUADRANT LOWER LEFT */
	0x259E, /*	QUADRANT UPPER RIGHT AND LOWER LEFT */
	0x258C, /*	LEFT HALF BLOCK */
	0x259B, /*	QUADRANT UPPER LEFT AND UPPER RIGHT AND LOWER LEFT */
	0x2584, /*	LOWER HALF BLOCK */
	0x259F, /*	QUADRANT UPPER RIGHT AND LOWER LEFT AND LOWER RIGHT */
	0x2599, /*	QUADRANT UPPER LEFT AND LOWER LEFT AND LOWER RIGHT */
	0x2588, /*	FULL BLOCK */
};
#endif

void qexpch(uchr cch, ushrt *n)
{
	ushrt n0;

	if (arch == ARCH_CPC)
	{
		qexpch_cpc(cch, n);
		return;
	}
	if (arch == ARCH_C64)
	{
		qexpch_c64(cch, n);
		return;
	}
	switch (cch)
	{
		case 0x0D:
		fprintf (outfile,"\n");
		xpos = 0;
		break;

		case '\\':fprintf (outfile,"\\\\");    ++xpos; break;
		case ';':
		case '/': if (xpos == 0) fputc('\\', outfile);
			  fputc(cch, outfile);
			  ++xpos;
		 	  break;

		case 0x10:	/* Ink */
		case 0x11:	/* Paper */
		case 0x12:	/* Flash */
		case 0x13:	/* Bright */
		case 0x14:	/* Inverse */
		case 0x15:	/* Over */
			  qnextbyte(n);	/* Skip control char and parameter */
			  break;
		case 0x16:	/* Position cursor */
		case 0x17:	/* Tabulation */
			  qnextbyte(n);	/* Skip control char and */
			  qnextbyte(n);	/* both parameters */
			  break;

		case 6:
		n0 = *n;
		/* Two htabs in succession are a line break */
		if (xpos < 15 && qnextbyte(&n0) == 6)
		{
			fprintf (outfile,"\n");
			xpos = 0;
			++(*n);
			break;
		}
		else if (xpos < 15)
		{
			fprintf (outfile,"\\0x09\\0x09");
			xpos = 16;
			break;
		}
		fprintf (outfile,"\n");
		xpos = 0;
		break;

	        default:
		if ((cch > 31) && (cch < 127)) 
		{
			fputc(cch,outfile);
			++xpos;
		}
		/* Block graphics / UDGs */
		else if (cch > 127 && cch < 165) 
		{	
			fprintf(outfile, "\\0x%02x", cch);
			++xpos;
		}
		else if (cch > 164) qexpdict (cch);
		else 
		{
			fprintf(outfile, "\\0x%02x", cch);
			++xpos;
		}
		if (xpos >= 32)
		{
			fprintf(outfile,"\n");
			xpos = 0;
		}
		break;
	}
}

#if 0
uchr mirror(uchr c)
{
	ushrt n;	
	uchr o;

	for (n = 0, o = 0; n < 8; n++)
	{
		o = o << 1;
		o |= (c & 1);
		c = c >> 1;
	}
	return o;
}

#endif

void qfontout(ushrt addr, ushrt nchars, char *title, int id)
{
	ushrt n;
	uchr b, m;

	fprintf(outfile,"<%s set=\"%d\">\n", title, id);

	for (n = 0; n < nchars * 8; n++)
	{
		b = zmem(addr + n);
		for (m = 0; m < 8; m++)
		{
			if (b & 0x80) fputc('#', outfile);
			else	      fputc('-', outfile);
			b = b << 1;	
		}
		fputc('\n', outfile);
	}
	fprintf(outfile,"</%s>\n", title);
}

#if 0
void qlistudgs(ushrt addr)
{
	fprintf(outfile, "  <udgs>\n");
	xfontout(addr, 21, "udg", 0);
	fprintf(outfile, "  </udgs>\n");
}

void qlistfont(ushrt addr)
{
	ushrt a, b, some, index, std_done;

	std_done = 0;
	some = index = 0;
/* [new in v0.9.0] */
/* Before doing a standard font dump, try searching the game code for writes
 * to the "font" system variable. This should bring out all fonts used by the
 * game engine, whether or not they are currently selected. */

/* This byte range is my guess at the area occupied by the Quill engine. */	
	fprintf(outfile, "  <fonts>\n");
	for (a = 0x5D00; a < 0x6C00; a++)
	{
/* This checks for a LD HL, immediately followed by a LD (5C36h),HL */
		if (zmem(a  ) == 0x21 && zmem(a+3) == 0x22 && 
		    zmem(a+4) == 0x36 && zmem(a+5) == 0x5C)
		{
			b = zword(a+1);
			++index;
			if (b < 0x5b00) 
			{
				fprintf(outfile, "  <!-- Font %d is the system font -->\n", index); 						
				continue;
			}
			++some;
			xfontout(b + 256, 96, "font", index);
			if (b == addr) std_done = 1;
		}
	}
/* If the search above didn't find the currently selected font, dump that. */
	if (!std_done)
	{
		if (addr >= 0x5B00) qfontout(addr+256, 96, "font", ++index);
		else fprintf(outfile, "  <!-- Font %d is the system font -->\n",
			 index); 						
	}	
	fprintf(outfile, "  </fonts>\n");
}


static const char *p1_moves[] = {"x=\"1\" y=\"0\"",
				 "x=\"1\" y=\"1\"",
				 "x=\"0\" y=\"1\"",
				 "x=\"-1\" y=\"1\"",
				 "x=\"-1\" y=\"0\"",
				 "x=\"-1\" y=\"-1\"",
				 "x=\"0\" y=\"-1\"",
				 "x=\"1\" y=\"-1\"" };

static const char *anshade[] = {"x", "y", "fill"};
static const char *angosub[] = {"no" };
static const char *anblock[] = {"h", "w", "x", "y" };

/* 2 is a lousy sample size, but all the Spectrum Quill games I've examined 
 * have their graphics tables at 0xFAB8. */

void qlistgfx(void)
{
	ushrt n, m;
	ushrt gfx_base  = 0xFAB8;	/* Arbitrary */
	ushrt gfx_ptrs  = zword(gfx_base);
	ushrt gfx_flags = zword(gfx_base + 2);
	ushrt gfx_count = zmem (gfx_base + 6);
	uchr gflag, nargs = 0, value;
	ushrt gptr;
	char inv, ovr;
	char opcode[30];
	ushrt neg[8];
	const char **argnames;

	fprintf(outfile, "  <graphics>\n");
	for (n = 0; n < gfx_count; n++)
	{
		fprintf(outfile, "    <graphic no=\"%d\"", n);

		gflag = zmem (gfx_flags + n);
		gptr  = zword(gfx_ptrs  + 2*n);

		if (gflag & 0x80) fprintf(outfile, " type=\"location\"");
		else		  fprintf(outfile, " type=\"sub\"");
		fprintf(outfile, " ink=\"%d\" paper=\"%d\" bit6=\"%d\">\n",
			(gflag & 7), (gflag >> 3) & 7, (gflag >> 6) & 1);

		do
		{
			memset(neg, 0, sizeof(neg));
			gflag = zmem(gptr);
			inv = ' '; ovr = ' ';
			if (gflag & 0x08) ovr = 'o';
			if (gflag & 0x10) inv = 'i';
			value = (gflag >> 3);
			opcode[0] = 0;
			argnames = anblock;
			switch(gflag & 7)
			{
				case 0:  nargs = 2;  
					 argnames = anshade;
                                         sprintf(opcode, "<plot");
					 if (ovr == 'o') 
						strcat(opcode, " over=\"1\"");
					 if (inv == 'i') 
						strcat(opcode, " inverse=\"1\"");
					 if (ovr == 'o' && inv == 'i')
						strcpy(opcode, "<amove");
					 break;
				case 1:  nargs = 2; 
					 argnames = anshade;
                                         if (gflag & 0x40) neg[0] = 1;
                                         if (gflag & 0x80) neg[1] = 1;

                                         sprintf(opcode, "<line");
					 if (ovr == 'o') 
						strcat(opcode, " over=\"1\"");
					 if (inv == 'i') 
						strcat(opcode, " inverse=\"1\"");
					 if (ovr == 'o' && inv == 'i')
						strcpy(opcode, "<move");	
					 break;
				case 2:  switch(gflag & 0x30)
					{
						case 0x00:
                                                	if (gflag & 0x40) neg[0] = 1;
                                                	if (gflag & 0x80) neg[1] = 1;
							nargs = 2; 
							strcpy(opcode, "<fill");
							argnames = anshade;
							break; 
						case 0x10:
							nargs = 4; 
							strcpy(opcode, "<block");
							argnames = anblock;
							break;
						case 0x20:
                                                	if (gflag & 0x40) neg[0] = 1;
                                                	if (gflag & 0x80) neg[1] = 1;
							nargs = 3; 
							argnames = anshade;
							strcpy(opcode, "<shade");
							break;
						case 0x30:
                                                	if (gflag & 0x40) neg[0] = 1;
                                                	if (gflag & 0x80) neg[1] = 1;
							nargs = 3; 
							argnames = anshade;
							strcpy(opcode, "<bshade");
							break;
						}
					 break;
				case 3:  nargs = 1; 
					 sprintf(opcode, "<gosub scale=\"%d\"",
						value & 7);
					 argnames = angosub;
					 break;
				case 4:  nargs = 0; 
					 sprintf(opcode, "<rplot");
					 if (ovr == 'o') 
						strcat(opcode, " over=\"1\"");
					 if (inv == 'i') 
						strcat(opcode, " inverse=\"1\"");
					 strcat(opcode, p1_moves[(value / 4)]);
					 break;
				case 5:  nargs = 0; 
					 if (gflag & 0x80) sprintf(opcode,
					  "<bright val=\"%d\"", value & 15);
					 else sprintf(opcode,
					  "<paper val=\"%d\"", value & 15);
					 break;
                                case 6:  nargs = 0; 
					 if (gflag & 0x80) sprintf(opcode,
					  "<flash val=\"%d\"", value & 15);
					 else sprintf(opcode,
					  "<ink val=\"%d\"", value & 15);
					 break;
				case 7:  nargs = 0; 
					 strcpy(opcode, "<end"); break;
			}
			fprintf(outfile, "      %s", opcode); 
			for (m = 0; m < nargs; m++)
			{
				fprintf(outfile, " %s=\"%s%d\"",
					argnames[m], 
					neg[m] ? "-" : "",
					zmem(gptr + 1 + m));
			}
			fprintf(outfile, " />\n");

			gptr += (nargs + 1);

		}
		while((gflag & 7) != 7);
		fprintf(outfile, "    </graphic>\n");
	}
	fprintf(outfile, "  </graphics>\n");
}

#endif

#endif
